function [] = GenerateSiftDescriptors( imageFileList, imageBaseDir, dataBaseDir, maxImageSize, gridSpacing, patchSize, canSkip )
%function [] = GenerateSiftDescriptors( imageFileList, imageBaseDir, dataBaseDir, maxImageSize, gridSpacing, patchSize, canSkip )
%
%Generate the dense grid of sift descriptors for each
% image
%
% imageFileList: cell of file paths
% imageBaseDir: the base directory for the image files
% dataBaseDir: the base directory for the data files that are generated
%  by the algorithm. If this dir is the same as imageBaseDir the files
%  will be generated in the same location as the image files
% maxImageSize: the max image size. If the image is larger it will be
%  resampeled.
% gridSpacing: the spacing for the grid to be used when generating the
%  sift descriptors
% patchSize: the patch size used for generating the sift descriptor
% canSkip: if true the calculation will be skipped if the appropriate data 
%  file is found in dataBaseDir. This is very useful if you just want to
%  update some of the data or if you've added new images.


if(nargin<4)
    maxImageSize = 1000
end

if(nargin<5)
    gridSpacing = 8
end

if(nargin<6)
    patchSize = 16
end

if(nargin<7)
    canSkip = 0
end



%%  compute histogram intersection kernel
fprintf('****************************************************************\n');
fprintf('Extract  local descriptors for every image per category\n');
fprintf('****************************************************************\n');

start_time = round(clock);

num_feature_files = size(imageFileList, 1);
parfor f = 1:size(imageFileList,1)
    if mod(f, 100) == 0
        fprintf('.');
    end
    
     [num_feature_files(f)  ~] = compute_features_one_image( imageFileList{f}, imageBaseDir, dataBaseDir, maxImageSize, gridSpacing, patchSize, canSkip);
end 
[num_feature_files(end)  dim_feature_files] = compute_features_one_image( imageFileList{end}, imageBaseDir, dataBaseDir, maxImageSize, gridSpacing, patchSize, 0);

outFName = fullfile(dataBaseDir, 'feature_summary.mat');
 if(size(dir(outFName),1)~=0 && canSkip)
     % do nothing
 else
     sp_make_dir(outFName);
     save(outFName, 'num_feature_files', 'dim_feature_files');
 end
fprintf('\n');

end_time = round(clock);
elapsed_time = end_time - start_time;
fprintf('Time elapsed: %5d days %5d hours %5d minutes %5d seconds\n\n', elapsed_time(3), elapsed_time(4), elapsed_time(5), elapsed_time(6));

end 

function [num_feature dim_feature] = compute_features_one_image( imageFName, imageBaseDir, dataBaseDir, maxImageSize, gridSpacing, patchSize, canSkip )
    % load image
    %imageFName = imageFileList{f};
    [dirN base] = fileparts(imageFName);
    baseFName = [dirN filesep base];
    outFName = fullfile(dataBaseDir, sprintf('%s_sift.mat', baseFName));
    imageFName = fullfile(imageBaseDir, imageFName);
    
    if(size(dir(outFName),1)~=0 && canSkip)
        %fprintf('Skipping %s\n', imageFName);
        num_feature = -1;
        dim_feature  = -1;
        return;
    end
      
    %I = imread(imageFName);
    I = sp_load_image(imageFName);

    [hgt wid depth] = size(I);
    if max(hgt,wid) > maxImageSize    %min->max  2013.7.31
        I = imresize(I, maxImageSize/max(hgt,wid), 'bicubic');  %min->max  2013.7.31
       % fprintf('Loaded %s: original size %d x %d, resizing to %d x %d\n',  imageFName, wid, hgt, size(I,2), size(I,1));
        [hgt wid depth] = size(I);
    end

%     %% find SIFT descriptors
%     % make grid (coordinates of upper left patch corners)
%     remX = mod(wid-patchSize,gridSpacing);
%     offsetX = floor(remX/2)+1;
%     remY = mod(hgt-patchSize,gridSpacing);
%     offsetY = floor(remY/2)+1;
%     [gridX,gridY] = meshgrid(offsetX:gridSpacing:wid-patchSize+1, offsetY:gridSpacing:hgt-patchSize+1);
%     siftArr = sp_find_sift_grid(I, gridX, gridY, patchSize, 0.8);
%     %siftArr = siftArr .^ 0.5;
%     siftArr = sp_normalize_sift(siftArr);
%     features.data = siftArr';
%     features.x = gridX(:) + patchSize/2 - 0.5;
%     features.y = gridY(:) + patchSize/2 - 0.5;
%     features.wid = wid;
%     features.hgt = hgt;
%     features.mask = ones(length(features.x), 1);
% %     fprintf('Processing %s: wid %d, hgt %d, grid size: %d x %d, %d patches\n',    imageFName, wid, hgt, size(gridX,2), size(gridX,1), numel(gridX));

    
    %%%%%%%%%%%%%%%
    %%2013.09.02
%            % Three scales
%             [siftArr sift_mask grid_x grid_y] = covcalculatornoncombination(I,patchSize+16, gridSpacing, 1,'MG');
%           %[siftArr, siftlen] = covcalculatornoncombination(I,16, 8, 1,'CM');
%             siftArr = reshape(siftArr, size(siftArr, 1) * size(siftArr, 2), size(siftArr, 3));
%             features.data = siftArr';
%             [gridX, gridY] = meshgrid(grid_x, grid_y);
%             features.x = gridX(:);
%             features.y = gridY(:);
%   
%             [siftArr sift_mask grid_x1 grid_y1] = covcalculatornoncombination(I,patchSize+8, gridSpacing, 1,'MG');
%             siftArr = siftArr(2:end-1, 2:end-1, :);
%             siftArr = reshape(siftArr, size(siftArr, 1) * size(siftArr, 2), size(siftArr, 3));
%             features.data = [features.data; siftArr'];
%              
%             [siftArr sift_mask grid_x grid_y] = covcalculatornoncombination(I,patchSize, gridSpacing, 1,'MG');
%             siftArr = siftArr(3:end-2, 3:end-2, :);
%             siftArr = reshape(siftArr, size(siftArr, 1) * size(siftArr, 2), size(siftArr, 3));
%             features.data = [features.data; siftArr'];

 % 2014.04.01           
            %[siftArr im_mask grid_x grid_y] = covcalculatornoncombination0(I,patchSize, gridSpacing, 1,'MG');  % MG CM
            [siftArr im_mask   grid_x grid_y] = covcalculatornoncombination(I,patchSize, gridSpacing,   1,'MG');  % MG CM
            %im_mask = and(im_mask, im_mask0);
            %siftArr = cat(3, siftArr0, siftArr1);
            
            siftArr = reshape(siftArr, size(siftArr, 1) * size(siftArr, 2), size(siftArr, 3));
            siftArr = sign(siftArr) .* abs(siftArr).^0.5;
            [gridX, gridY] = meshgrid(grid_x, grid_y);
            features.x = gridX(:);
            features.y = gridY(:);
            features.mask = im_mask(:);
            features.data =siftArr;
            features.data = bsxfun(@rdivide, features.data, max(sqrt(sum(features.data .^2, 2)), eps(2)));
            %features.data(:,         1:size(siftArr0, 3))= bsxfun(@rdivide, features.data(:, 1:size(siftArr0, 3)),         max(sqrt(sum(features.data(:, 1:size(siftArr0, 3)) .^2, 2)),        eps(2)));
            %features.data(:, size(siftArr0, 3)+1:end) = bsxfun(@rdivide, features.data(:, size(siftArr0, 3)+1:end), max(sqrt(sum(features.data(:, size(siftArr0, 3)+1:end) .^2, 2)), eps(2)));
            %features.data = bsxfun(@rdivide, features.data, max(sqrt(sum(features.data .^2, 2)), eps(2)));
            features.data = features.data';
            features.wid = wid;
            features.hgt = hgt;
       
            num_feature = size(features.data, 2);
            dim_feature = size(features.data, 1);

    sp_make_dir(outFName);
    save(outFName, 'features');

end




